//------------------------------------------------------------------------------
//
//   Greg's Blocks - BlockShape
//
//------------------------------------------------------------------------------

package gcewing.blocks;

import java.util.List;

import net.minecraft.block.*;
import net.minecraft.block.material.*;
import net.minecraft.client.renderer.texture.*;
import net.minecraft.entity.*;
import net.minecraft.entity.player.*;
import net.minecraft.init.*;
import net.minecraft.item.*;
import net.minecraft.tileentity.*;
import net.minecraft.world.*;
import net.minecraft.util.*;

public class BlockShape extends BlockContainer {

	public static int modelID;
	public static TEShape lastPlacedTileEntity;

	public BlockShape() {
		super(Material.rock);
	}

	@Override
	public void registerBlockIcons(IIconRegister reg) {
	}

	@Override
	public void onBlockPlacedBy(World world, int x, int y, int z, EntityLivingBase player, ItemStack stack) {
		int angle = Utils.placementRotation(player);
		world.setBlockMetadataWithNotify(x, y, z, angle, 0x3);
		lastPlacedTileEntity = (TEShape)world.getTileEntity(x, y, z);
		//System.out.printf("BlockShape.onBlockPlacedBy: tile entity = %s\n", lastPlacedTileEntity);
	}
	
	@Override
	public TileEntity createNewTileEntity(World world, int meta) {
		return new TEShape();
	}

	@Override
	public boolean renderAsNormalBlock() {
		return false;
	}
	
	@Override
	public int getRenderType() {
		return modelID;
	}
	
	@Override
	public boolean isOpaqueCube() {
		return false;
	}
	
//	@Override
//	public IIcon getBlockTexture(IBlockAccess world, int x, int y, int z, int side) {
//		System.out.printf("BlockShape.getBlockTexture\n");
//		TileEntity te = world.getTileEntity(x, y, z);
//		if (te instanceof TEShape)
//			return getBaseBlockTexture((TEShape)te);
//		else
//			return null;
//	}
//
//	IIcon getBaseBlockTexture(TEShape te) {
//		return te.getBaseBlock().getBlockTextureFromSideAndMetadata(0, te.baseMetadata);
//	}

	@Override
	public IIcon getIcon(int par1, int par2) {
		// This is needed for the particle effects when the block is broken.
		// Don't get the world coordinates, unfortunately.
		//System.out.printf("BlockShape.getBlockTextureFromSideAndMetadata\n");
		return Blocks.stone.getIcon(0, 0);
	}
	
	@Override
	public void addCollisionBoxesToList(World world, int x, int y, int z, 
		AxisAlignedBB clip, List list, Entity entity)
	{
		int data = world.getBlockMetadata(x, y, z);
		TEShape te = (TEShape)world.getTileEntity(x, y, z);
		Matrix3 rot = Matrix3.blockRotations[data];
		Vector3 org = new Vector3(x + 0.5, y + 0.5, z + 0.5);
		addBox(-0.5, 0.5,   -0.5, 0.0,  -0.5, 0.5,   rot, org, clip, list);
		if (te.shape == 0 || te.shape == 2)
			addBox(-0.5, 0.5,   0.0, 0.5,   0.0, 0.5,   rot, org, clip, list);
		if (te.shape == 1)
			addBox(-0.5, 0.0,   0.0, 0.5,   0.0, 0.5,   rot, org, clip, list);
		if (te.shape == 2)
			addBox(-0.5, 0.0,   0.0, 0.5,   -0.5, 0.0,  rot, org, clip, list);
	}

	void addBox(double x0, double x1, double y0, double y1, double z0, double z1,
			Matrix3 rot, Vector3 org, AxisAlignedBB clip, List list)
	{
		Vector3 p0 = rot.mul(x0, y0, z0).add(org);
		Vector3 p1 = rot.mul(x1, y1, z1).add(org);
		if (p0.x < p1.x) {x0 = p0.x; x1 = p1.x;} else {x0 = p1.x; x1 = p0.x;}
		if (p0.y < p1.y) {y0 = p0.y; y1 = p1.y;} else {y0 = p1.y; y1 = p0.y;}
		if (p0.z < p1.z) {z0 = p0.z; z1 = p1.z;} else {z0 = p1.z; z1 = p0.z;}
		AxisAlignedBB box = AxisAlignedBB.getAABBPool().getAABB(x0, y0, z0, x1, y1, z1);
		//System.out.printf("BlockShape.addBox: %s\n", box);
		if (box != null && clip.intersectsWith(box))
			list.add(box);
	}
	
	public boolean canHarvestBlock(EntityPlayer player, int meta) {
		// Item drop occurs in onBlockRemoval
		System.out.printf("BlockShape.canHarvestBlock: by %s\n", player);
		return false;
	}

//	public ArrayList<ItemStack> getBlockDropped(World world, int x, int y, int z, int metadata, int fortune) {
//		//ArrayList<ItemStack> result = super.getBlockDropped(world, x, y, z, metadata, fortune);
//		ArrayList<ItemStack> result = new ArrayList<ItemStack>();
//		ItemStack stack = new ItemStack(this, 1);
//		TileEntity te = world.getTileEntity(x, y, z);
//		stack.setTagCompound(Utils.nbtFromTileEntity(te));
//		result.add(stack);
//		System.out.printf("BlockShape.getBlockDropped --> %s\n", result);
//		return result;
//	}

	@Override
	public void onBlockHarvested(World world, int x, int y, int z, int data, EntityPlayer player) {
		//System.out.printf("BlockShape.onBlockHarvested: player = %s\n", player);
		if (!Utils.playerIsInCreativeMode(player)) {
			TileEntity te = world.getTileEntity(x, y, z);
			ItemStack stack = Utils.blockStackWithTileEntity(this, 1, te);
			dropBlockAsItem(world, x, y, z, stack);
		}
	}

//	public float blockStrength(World world, EntityPlayer player, int x, int y, int z) {

	@Override
	public float getPlayerRelativeBlockHardness(EntityPlayer player, World world, int x, int y, int z) {
		TEShape te = (TEShape)world.getTileEntity(x, y, z);
		if (te != null) {
			Block base = te.getBaseBlock();
			if (base != null)
				return base.getPlayerRelativeBlockHardness(player, world, x, y, z);
		}
		return 0.0F;
	}

	@Override
	public void setBlockBoundsBasedOnState(IBlockAccess world, int x, int y, int z) {
		TEShape te = (TEShape)world.getTileEntity(x, y, z);
		if (te.shape <= 2)
			setBlockBounds(0.0F, 0.0F, 0.0F, 1.0F, 1.0F, 1.0F);
		else
			setBlockBounds(0.0F, 0.0F, 0.0F, 1.0F, 0.5F, 1.0F);
	}

}
