//------------------------------------------------------------------------------
//
//   Greg's Blocks - Renderer for shaped blocks
//
//------------------------------------------------------------------------------

package gcewing.blocks;

import java.nio.*;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.glu.GLU;
import net.minecraft.client.Minecraft;

import net.minecraft.block.*;
import net.minecraft.client.renderer.*;
import net.minecraft.tileentity.*;
import net.minecraft.world.*;
import net.minecraft.util.*;

import net.minecraftforge.client.*;
import cpw.mods.fml.client.registry.ISimpleBlockRenderingHandler;

public class RenderShape implements ISimpleBlockRenderingHandler {

	final static int ridgeShapes[] = {TEShape.ridgeShape, TEShape.smartRidgeShape};
	final static int ridgeOrSlopeShapes[] = {TEShape.ridgeShape, TEShape.smartRidgeShape,
		TEShape.roofShape, TEShape.outerCornerShape, TEShape.innerCornerShape};
	final static int valleyShapes[] = {TEShape.valleyShape, TEShape.smartValleyShape};
	final static int valleyOrSlopeShapes[] = {TEShape.valleyShape, TEShape.smartValleyShape,
		TEShape.roofShape, TEShape.innerCornerShape};

	static Tessellator tess;
	static IBlockAccess blockWorld;
	static double blockX, blockY, blockZ; // world coords of block centre
	static double wx, wy, wz; // vertex transformed to world coords
	static double tu, tv; // vertex uv coords
	static Matrix3 rotation;
	static double tx0, ty0, tw, th; // base texture uv origin and size
	static float cmr, cmg, cmb; // colour multiplier
	static Block baseBlock;
	static int blockBrightness;
	static boolean ao; // using ambient occlusion
	static Vector3 faceU, faceV, faceN; // transformed face plane basis and normal
	static int faceBrightness;
	static float faceLightValue;

	@Override
	public int getRenderId() {
		return BlockShape.modelID;
	}

	@Override
	public void renderInventoryBlock(Block block, int metadata, int modelID, RenderBlocks renderer) {
		// Can't use this because we need the ItemStack's NBTTagCompound
	}

	@Override
	public boolean renderWorldBlock(IBlockAccess world, int x, int y, int z, 
		Block block, int modelId, RenderBlocks render)
	{
		int data = world.getBlockMetadata(x, y, z);
		TileEntity te = world.getTileEntity(x, y, z);
		RenderShape.render(world, block, data, (TEShape)te, x, y, z, render.overrideBlockTexture);
		return true;
	}
	
	@Override
	public boolean shouldRender3DInInventory(int meta) {
		return true; //???
	}
	
	public static void render(IBlockAccess world, Block block, int data, TEShape te,
		int x, int y, int z, IIcon texIcon)
	{
		//if (x != 0 || y != 0 || z != 0)
		//	System.out.printf("RenderShape.render: %d %d %d data %d shape %d base %d:%d\n",
		//		x, y, z, data, te.shape, te.baseBlockID, te.baseMetadata);
		baseBlock = te.getBaseBlock();
		if (baseBlock == null)
			return;
		String texture;
		blockWorld = world;
		blockX = x + 0.5;
		blockY = y + 0.5;
		blockZ = z + 0.5;
		rotation = Matrix3.blockRotations[data & 3];
		if (texIcon == null)
			texIcon = baseBlock.getIcon(2, te.baseMetadata);
		tx0 = texIcon.getMinU();
		ty0 = texIcon.getMinV();
		tw = texIcon.getMaxU() - tx0;
		th = texIcon.getMaxV() - ty0;
//		System.out.printf("RenderShape.render: tx0 = %.4f, ty0 = %.4f, tw = %.4f, th = %.4f\n",
//			tx0, ty0, tw, th);
		tess = Tessellator.instance;
		if (world != null) {
			blockBrightness = block.getMixedBrightnessForBlock(world, x, y, z);
    	setColorMultiplier(baseBlock.colorMultiplier(world, x, y, z));
    }
    else {
    	blockBrightness = 0xf000f0;
    	setColorMultiplier(0xffffff);
    }
    ao = Minecraft.isAmbientOcclusionEnabled();
		switch (te.shape) {
			case TEShape.roofShape: renderSlope(); break;
			case TEShape.outerCornerShape: renderOuterCorner(); break;
			case TEShape.innerCornerShape: renderInnerCorner(); break;
			case TEShape.ridgeShape: renderRidge(); break;
			case TEShape.smartRidgeShape: renderSmartRidge(); break;
			case TEShape.valleyShape: renderValley(); break;
			case TEShape.smartValleyShape: renderSmartValley(); break;
		}
//		ForgeHooksClient.unbindTexture();
	}
	
	static void setColorMultiplier(int color) {
		//System.out.printf("RenderShape.setColorMultiplier: 0x%08x\n", color);
		//cmr = cmg = cmb = 1.0F;
		cmr = ((color >> 16) & 0xff) / 255.0F;
		cmg = ((color >> 8) & 0xff) / 255.0F;
		cmb = (color & 0xff) / 255.0F;
	}
	
	static void beginTopFace() {
		beginOuterFace(Vector3.unitZ, Vector3.unitX, Vector3.unitY);
	}
	
	static void beginBottomFace() {
		beginOuterFace(Vector3.unitX, Vector3.unitZ, Vector3.unitNY);
	}
	
	static void beginPosXFace() {
		beginOuterFace(Vector3.unitY, Vector3.unitZ, Vector3.unitX);
	}
	
	static void beginNegXFace() {
		beginOuterFace(Vector3.unitZ, Vector3.unitY, Vector3.unitNX);
	}
	
	static void beginPosZFace() {
		beginOuterFace(Vector3.unitX, Vector3.unitY, Vector3.unitZ);
	}
	
	static void beginNegZFace() {
		beginOuterFace(Vector3.unitY, Vector3.unitX, Vector3.unitNZ);
	}
	
	static void beginPosXSlope() {
		beginInnerFace(Vector3.unitPXPY);
	}
	
	static void beginNegXSlope() {
		beginInnerFace(Vector3.unitNXPY);
	}
	
	static void beginPosZSlope() {
		beginInnerFace(Vector3.unitPYPZ);
	}
	
	static void beginNegZSlope() {
		beginInnerFace(Vector3.unitPYNZ);
	}
	
	static void beginInnerFace(Vector3 n) {
		// n = normal
		beginFace(n);
		lightFace(faceLightValue, blockBrightness);
	}
	
	static void beginOuterFace(Vector3 u, Vector3 v, Vector3 n) {
		// u, v = face plane basis, n = normal
		beginFace(n);
		if (blockWorld == null)
			shadeFace(faceLightValue);
		else if (ao) {
			faceU = rotation.mul(u);
			faceV = rotation.mul(v);
		}
		else
			lightFace(faceLightValue, faceBrightness);
	}
	
	static void beginFace(Vector3 n) {
		faceN = rotation.mul(n);
		normal(faceN.x, faceN.y, faceN.z);
		Vector3 q = faceN.add(blockX, blockY, blockZ);
		int qx = (int)Math.floor(q.x);
		int qy = (int)Math.floor(q.y);
		int qz = (int)Math.floor(q.z);
		if (blockWorld != null)
			faceBrightness = baseBlock.getMixedBrightnessForBlock(blockWorld, qx, qy, qz);
		double nx = faceN.x * faceN.x;
		double ny = faceN.y * faceN.y;
		double nz = faceN.z * faceN.z;
		faceLightValue = (float)(0.6 * nx * nx + 0.8 * nz + (faceN.y > 0 ? 1 : 0.5) * ny);
	}
	
	static void normal(double x, double y, double z) {
		Vector3 v = rotation.mul(x, y, z);
		tess.setNormal((float)v.x, (float)v.y, (float)v.z);
	}
	
	static void corner(double x, double y, double z, double tx, double ty) {
		cornerOrInteriorVertex(x, y, z, tx, ty, true);
	}
	
	static void vertex(double x, double y, double z, double tx, double ty) {
		cornerOrInteriorVertex(x, y, z, tx, ty, false);
	}

	static void cornerOrInteriorVertex(double x, double y, double z, double tx, double ty, boolean corner) {
		Vector3 p = rotation.mul(x - 0.5, y - 0.5, z - 0.5);
		wx = blockX + p.x;
		wy = blockY + p.y;
		wz = blockZ + p.z;
		if (blockWorld != null) {
			if (ao) {
				if (corner)
					lightCornerVertex(wx, wy, wz);
				else
					lightInteriorVertex(wx, wy, wz);
			}
		}
//		tu = tx0 + tx / 16.0;
//		tv = ty0 + ty / 16.0;
		tu = tx0 + tx * tw;
		tv = ty0 + ty * th;
		tess.addVertexWithUV(wx, wy, wz, tu, tv);
	}
	
	static void dummy() {
		// Dummy fourth vertex for a triangular face
		tess.addVertexWithUV(wx, wy, wz, tu, tv);
	}
	
	static void lightCornerVertex(double x, double y, double z) {
		int aoBrightness, aoBrightnessSum = 0;
		float aoLightValueSum = 0;
		Vector3 p = faceN.mul(0.5).add(x, y, z);
		for (int i = 0; i <= 1; i++)
			for (int j = 0; j <= 1; j++) {
				Vector3 q = p.add(faceU.mul(i - 0.5)).add(faceV.mul(j - 0.5));
				int qx = (int)Math.floor(q.x);
				int qy = (int)Math.floor(q.y);
				int qz = (int)Math.floor(q.z);
				aoBrightness = baseBlock.getMixedBrightnessForBlock(blockWorld, qx, qy, qz);
				if (aoBrightness == 0)
					aoBrightness = faceBrightness;
				aoBrightnessSum += aoBrightness;
				aoLightValueSum += blockWorld.getBlock(qx, qy, qz).getAmbientOcclusionLightValue();
			}
		lightFace(faceLightValue * aoLightValueSum / 4, (aoBrightnessSum >> 2) & 0xff00ff); 
	}
	
	static void lightInteriorVertex(double xv, double yv, double zv) {
		int br = 0, brSum = 0, brCount = 0;
		for (int i = 0; i <= 1; i++)
			for (int j = 0; j <= 1; j++)
				for (int k = 0; k <= 1; k++) {
					int xb = (int)Math.floor(xv - 0.5 + i);
					int yb = (int)Math.floor(yv - 0.5 + j);
					int zb = (int)Math.floor(zv - 0.5 + k);
					br = baseBlock.getMixedBrightnessForBlock(blockWorld, xb, yb, zb);
					if (br != 0) {
						brSum += br;
						brCount += 1;
					}
				}
		if (brCount > 0) {
			int br1 = (brSum >> 16) / brCount;
			int br2 = (brSum & 0xffff) / brCount;
			br = (br1 << 16) | br2;
		}
		lightFace(faceLightValue, br);
	}

	static void lightFace(float bm, int br) {
		//System.out.printf("RenderShape.lightFace: %s, 0x%x\n", bm, br);
		tess.setBrightness(br);
		shadeFace(bm);
	}
	
	static void shadeFace(float bm) {
		//System.out.printf("RenderShape.shadeFace: %s\n", bm);
		tess.setColorOpaque_F(bm * cmr, bm * cmg, bm * cmb);
	}
	
	static boolean hasNeighbour(int dx, int dy, int dz, int[] shapes) {
		TileEntity te = getNeighbouringTileEntity(dx, dy, dz);
		if (te instanceof TEShape) {
			TEShape ste = (TEShape)te;
			int shape = ste.shape;
			//if (blockWorld != null)
			//	System.out.printf("RenderShape.neighbouringRidgeAt(%d,%d,%d): shape = %d\n",
			//		dx, dy, dz, shape);
			//return shape == TEShape.ridgeShape || shape == TEShape.smartRidgeShape;
			for (int i = 0; i < shapes.length; i++)
				if (shape == shapes[i])
					return true;
		}
		else {
			//if (blockWorld != null)
			//	System.out.printf("RenderShape.neighbouringRidgeAt(%d,%d,%d): Not a shape\n",
			//		dx, dy, dz);
		}
		return false;
	}
	
	static boolean ridgeAt(int dx, int dy, int dz) {
		return hasNeighbour(dx, dy, dz, ridgeShapes);
	}
	
	static boolean ridgeOrSlopeAt(int dx, int dy, int dz) {
		return hasNeighbour(dx, dy, dz, ridgeOrSlopeShapes);
	}
	
	static boolean valleyAt(int dx, int dy, int dz) {
		return hasNeighbour(dx, dy, dz, valleyShapes);
	}
	
	static boolean valleyOrSlopeAt(int dx, int dy, int dz) {
		return hasNeighbour(dx, dy, dz, valleyOrSlopeShapes);
	}
	
	static TileEntity getNeighbouringTileEntity(int dx, int dy, int dz) {
		if (blockWorld != null) {
			Vector3 p = rotation.mul(dx, dy, dz);
			//System.out.printf("getNeighbouringTileEntity: (%s,%s,%s) + (%s,%s,%s)\n",
			//	blockX, blockY, blockZ, p.x, p.y, p.z);
			int x = (int)Math.floor(blockX + p.x);
			int y = (int)Math.floor(blockY + p.y);
			int z = (int)Math.floor(blockZ + p.z);
			return blockWorld.getTileEntity(x, y, z);
		}
		else
			return null;
	}
	
	//------------------------------------------------------------------------------
	
	static void leftQuad() {
		beginPosXFace();
		corner(1, 1, 1,   0, 0);
		corner(1, 0, 1,   0, 1);
		corner(1, 0, 0,   1, 1);
		corner(1, 1, 0,   1, 0);
	}
	
	static void rightQuad() {
		beginNegXFace();
		corner(0, 1, 0,   0, 0);
		corner(0, 0, 0,   0, 1);
		corner(0, 0, 1,   1, 1);
		corner(0, 1, 1,   1, 0);
	}
	
	static void frontQuad() {
		beginNegZFace();
		corner(1, 1, 0,   0, 0);
		corner(1, 0, 0,   0, 1);
		corner(0, 0, 0,   1, 1);
		corner(0, 1, 0,   1, 0);
	}

	static void backQuad() {
		beginPosZFace();
		corner(0, 1, 1,   0, 0);
		corner(0, 0, 1,   0, 1);
		corner(1, 0, 1,   1, 1);
		corner(1, 1, 1,   1, 0);
	}

	static void bottomQuad() {
		beginBottomFace();
		corner(0, 0, 1,   0, 0);
		corner(0, 0, 0,   0, 1);
		corner(1, 0, 0,   1, 1);
		corner(1, 0, 1,   1, 0);
	}
	
	static void leftTriangle() {
		beginPosXFace();
		corner(1, 1, 1,   0, 0);
		corner(1, 0, 1,   0, 1);
		corner(1, 0, 0,   1, 1);
		corner(1, 1, 1,   0, 0);
	}
	
	static void rightTriangle() {
		beginNegXFace();
		corner(0, 1, 1,   1, 0);
		corner(0, 0, 0,   0, 1);
		corner(0, 0, 1,   1, 1);
		dummy();
	}
	
	static void ridgeLeftFace() {
		beginPosXFace();
		vertex(1, 0.5, 0.5,   0.5, 0.5);
		corner(1, 0,   1,     0,   1);
		corner(1, 0,   0,     1,   1);
		dummy();
	}
		
	static void ridgeRightFace() {
		beginNegXFace();
		vertex(0, 0.5, 0.5,   0.5, 0.5);
		corner(0, 0,   0,     0,   1  );
		corner(0, 0,   1,     1,   1  );
		dummy();
	}
	
	static void ridgeBackFace() {
		beginPosZFace();
		vertex(0.5, 0.5, 1,   0.5, 0.5);
		corner(0,   0,   1,   0,   1  );
		corner(1,   0,   1,   1,   1  );
		dummy();
	}
	
	static void ridgeFrontFace() {
		beginNegZFace();
		vertex(0.5, 0.5, 0,   0.5, 0.5);
		corner(1,   0,   0,   0,   1  );
		corner(0,   0,   0,   1,   1  );
		dummy();
	}
	
	static void ridgeFrontSlope() {
		beginNegZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   0,   0,     0,   1  );
		vertex(0,   0,   0,     1,   1  );
		dummy();
	}
	
//	static void ridgeFrontLeftCorner() {
//		beginNegZSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(1,   0.5, 0.5,   0,   0.5);
//		vertex(1,   0,   0,     0,   1  );
//		dummy();
//		beginPosXSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(1,   0,   0,     1,   1  );
//		vertex(0.5, 0.5, 0,     1,   0.5);
//		dummy();
//	}
	
//	static void ridgeFrontRightCorner() {
//		beginNegXSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0.5, 0.5, 0,     0,   0.5);
//		vertex(0,   0,   0,     0,   1);
//		dummy();
//		beginNegZSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0,   0,   0,     1,   1  );
//		vertex(0,   0.5, 0.5,   1,   0.5);
//		dummy();
//	}
	
//	static void ridgeBackLeftCorner() {
//		beginPosXSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0.5, 0.5, 1,     0,   0.5);
//		vertex(1,   0,   1,     0,   1  );
//		dummy();
//		beginPosZSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(1,   0,   1,     1,   1  );
//		vertex(1,   0.5, 0.5,   1,   0.5);
//		dummy();
//	}
	
//	static void ridgeBackRightCorner() {
//		beginNegXSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0,   0,   1,     1,   1  );
//		vertex(0.5, 0.5, 1,     1,   0.5);
//		dummy();
//		beginPosZSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0,   0.5, 0.5,   0,   0.5);
//		vertex(0,   0,   1,     0,   1  );
//		dummy();
//	}
	
	static void ridgeBackSlope() {
		beginPosZSlope();
		vertex(0, 0.5, 0.5,   0, 0.5);
		vertex(0, 0,   1,     0, 1  );
		vertex(1, 0,   1,     1, 1  );
		vertex(1, 0.5, 0.5,   1, 0.5);
	}
	
	static void ridgeLeft() {
		if (ridgeOrSlopeAt(1, 0, 0))
			connectRidgeLeft();
		else {
			beginPosXSlope();
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			vertex(1,   0,   1,     0,   1);
			vertex(1,   0,   0,     1,   1);
			dummy();
		}
	}
	
	static void connectRidgeLeft() {
		beginNegZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   0.5, 0.5,   0,   0.5);
		vertex(1,   0,   0,     0,   1  );
		dummy();
		beginPosZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   0,   1,     1,   1  );
		vertex(1,   0.5, 0.5,   1,   0.5);
		dummy();
	}

	static void ridgeRight() {
		if (ridgeOrSlopeAt(-1, 0, 0))
			connectRidgeRight();
		else {
			beginNegXSlope();
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			vertex(0,   0,   0,     0,   1);
			vertex(0,   0,   1,     1,   1);
			dummy();
		}
	}
	
	static void connectRidgeRight() {
		beginNegZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   0,   0,     1,   1  );
		vertex(0,   0.5, 0.5,   1,   0.5);
		dummy();
		beginPosZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   0.5, 0.5,   0,   0.5);
		vertex(0,   0,   1,     0,   1  );
		dummy();
	}
	
	static void ridgeFront() {
		if (ridgeOrSlopeAt(0, 0, -1))
			connectRidgeFront();
		else {
			beginNegZSlope();
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			vertex(1,   0,   0,     0,   1);
			vertex(0,   0,   0,     1,   1);
			dummy();
		}
	}
	
	static void connectRidgeFront() {
		beginPosXSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   0,   0,     1,   1  );
		vertex(0.5, 0.5, 0,     1,   0.5);
		dummy();
		beginNegXSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0.5, 0.5, 0,     0,   0.5);
		vertex(0,   0,   0,     0,   1);
		dummy();
	}
	
	static void ridgeBack() {
		if (ridgeOrSlopeAt(0, 0,  1))
			connectRidgeBack();
		else {
			beginPosZSlope();
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			vertex(0,   0,   1,     0,   1);
			vertex(1,   0,   1,     1,   1);
			dummy();
		}
	}
	
	static void connectRidgeBack() {
		beginPosXSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0.5, 0.5, 1,     0,   0.5);
		vertex(1,   0,   1,     0,   1  );
		dummy();
		beginNegXSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   0,   1,     1,   1  );
		vertex(0.5, 0.5, 1,     1,   0.5);
		dummy();
	}
	
	static void connectValleyLeft() {
		beginPosZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1, 0.5, 0.5,   1, 0.5);
		vertex(1, 1, 0,   1, 0);
		dummy();
		beginNegZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1, 1, 1,   0, 0);
		vertex(1, 0.5, 0.5,   0, 0.5);
		dummy();
		valleyEndLeft();
	}
	
	static void connectValleyRight() {
		beginPosZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   1,   0,     0,   0);
		vertex(0,   0.5, 0.5,   0, 0.5);
		dummy();
		beginNegZSlope();
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   0.5, 0.5,   1,   0.5);
		vertex(0,   1,   1,     1,   0);
		dummy();
		valleyEndRight();
	}
	
	static void connectValleyFront() {
		beginPosXSlope();
		vertex(0,   1,   0,     1,   0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0.5, 0.5, 0,     1,   0.5);
		dummy();
		beginNegXSlope();
		vertex(1,   1,   0,     0,   0);
		vertex(0.5, 0.5, 0,     0,   0.5);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		dummy();
		valleyEndFront();
	}
	
	static void connectValleyBack() {
		beginPosXSlope();
		vertex(0,   1,   1,     0,   0);
		vertex(0.5, 0.5, 1,     0,   0.5);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		dummy();
		beginNegXSlope();
		vertex(1,   1,   1,     1,   0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0.5, 0.5, 1,     1,   0.5);
		dummy();
		valleyEndBack();
	}
	
	static void valleyEndLeft() {
		beginPosXFace();
		corner(1, 1,   1,     0,   0);
		corner(1, 0,   1,     0,   1);
		vertex(1, 0.5, 0.5,   0.5, 0.5);
		dummy();
		corner(1, 0,   1,     0,   1);
		corner(1, 0,   0,     1,   1);
		vertex(1, 0.5, 0.5,   0.5, 0.5);
		dummy();
		corner(1, 0,   0,     1,   1);
		corner(1, 1,   0,     1,   0);
		vertex(1, 0.5, 0.5,   0.5, 0.5);
		dummy();
	}
	
	static void valleyEndRight() {
		beginNegXFace();
		corner(0, 0,   1,     1,   1);
		corner(0, 1,   1,     1,   0);
		vertex(0, 0.5, 0.5,   0.5, 0.5);
		dummy();
		corner(0, 0,   0,     0,   1);
		corner(0, 0,   1,     1,   1);
		vertex(0, 0.5, 0.5,   0.5, 0.5);
		dummy();
		corner(0, 1,   0,     0,   0);
		corner(0, 0,   0,     0,   1);
		vertex(0, 0.5, 0.5,   0.5, 0.5);
		dummy();
	}
	
	static void valleyEndFront() {
		beginNegZFace();
		corner(1,   1,   0,    0,   0);
		corner(1,   0,   0,    0,   1);
		vertex(0.5, 0.5, 0,    0.5, 0.5);
		dummy();
		corner(1,   0,   0,    0,   1);
		corner(1,   0,   0,    1,   1);
		vertex(0.5, 0.5, 0,    0.5, 0.5);
		dummy();
		corner(0,   0,   0,    1,   1);
		corner(0,   1,   0,    1,   0);
		vertex(0.5, 0.5, 0,    0.5, 0.5);
		dummy();
	}
	
	static void valleyEndBack() {
		beginPosZFace();
		corner(0,   1,   1,     0,   0);
		corner(0,   0,   1,     0,   1);
		vertex(0.5, 0.5, 1,     0.5, 0.5);
		dummy();
		corner(0,   0,   1,     0,   1);
		corner(1,   0,   1,     1,   1);
		vertex(0.5, 0.5, 1,     0.5, 0.5);
		dummy();
		corner(1,   0,   1,     1,   1);
		corner(1,   1,   1,     1,   0);
		vertex(0.5, 0.5, 1,     0.5, 0.5);
		dummy();
	}
	
	static void smartValleyLeft() {
		if (valleyOrSlopeAt(1, 0, 0))
			connectValleyLeft();
		else
			terminateValleyLeft();
	}
	
	static void terminateValleyLeft() {
		beginNegXSlope();
		vertex(1,   1,   0,     0,   0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   1,    1,    1,   0);
		dummy();
		leftQuad();
	}
	
	static void smartValleyRight() {
		if (valleyOrSlopeAt(-1, 0, 0))
			connectValleyRight();
		else
			terminateValleyRight();
	}
	
	static void terminateValleyRight() {
		beginPosXSlope();
		vertex(0, 1, 1,   0, 0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0, 1, 0,   1, 0);
		dummy();
		rightQuad();
	}
	
	static void smartValleyFront() {
		if (valleyOrSlopeAt(0, 0, -1))
			connectValleyFront();
		else
			terminateValleyFront();
	}
	
	static void terminateValleyFront() {
		beginPosZSlope();
		vertex(0,   1,   0,     0,   0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1,   1,   0,     1,   0);
		dummy();
		frontQuad();
	}
	
	static void smartValleyBack() {
		if (valleyOrSlopeAt(0, 0, 1))
			connectValleyBack();
		else
			terminateValleyBack();
	}
	
	static void terminateValleyBack() {
		beginNegZSlope();
		vertex(1,   1,   1,     0,   0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(0,   1,   1,     1,   0);
		dummy();
		backQuad();
	}
	
	//------------------------------------------------------------------------------

	static void renderSlope() {
		// Sloping face
		beginNegZSlope();
		if (valleyAt(0, 0, 1)) {
			vertex(1, 1, 1,   0, 0);
			vertex(1, 0, 0,   0, 1);
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			dummy();
			vertex(1, 0, 0,   0, 1);
			vertex(0, 0, 0,   1, 1);
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			dummy();
			vertex(0, 0, 0,   1, 1);
			vertex(0, 1, 1,   1, 0);
			vertex(0.5, 0.5, 0.5,   0.5, 0.5);
			dummy();
			connectValleyBack();
		}
		else {
			vertex(1, 1, 1,   0, 0);
			vertex(1, 0, 0,   0, 1);
			vertex(0, 0, 0,   1, 1);
			vertex(0, 1, 1,   1, 0);
			backQuad();
		}
		// Other faces
		leftTriangle();
		rightTriangle();
		bottomQuad();
		if (ridgeAt(0, 0, -1))
			connectRidgeFront();
	}
	
	static void renderOuterCorner() {
		// Front slope
		beginNegZSlope();
		vertex(0, 1, 1,   1, 0);
		vertex(1, 0, 0,   0, 1);
		vertex(0, 0, 0,   1, 1);
		dummy();
		// Left slope
		beginPosXSlope();
		vertex(0, 1, 1,   0, 0);
		vertex(1, 0, 1,   0, 1);
		vertex(1, 0, 0,   1, 1);
		dummy();
		// Back
		beginNegZFace();
		corner(0, 1, 1,   0, 0);
		corner(0, 0, 1,   0, 1);
		corner(1, 0, 1,   1, 1);
		dummy();
		// Other faces
		rightTriangle();
		bottomQuad();
		if (ridgeAt(0, 0, -1))
			connectRidgeFront();
		if (ridgeAt(1, 0, 0))
			connectRidgeLeft();
	}
	
	static void renderInnerCorner() {
		// Left slope
		beginPosXSlope();
		vertex(0, 1, 0,   1, 0);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		vertex(1, 0, 0,   1, 1);
		dummy();
		if (valleyAt(-1, 0, 0))
			connectValleyRight();
		else
			terminateValleyRight();
		// Front slope
		beginNegZSlope();
		vertex(1, 1, 1,   0, 0);
		vertex(1, 0, 0,   0, 1);
		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
		dummy();
		if (valleyAt(0, 0, 1))
			connectValleyBack();
		else
			terminateValleyBack();
		// Front triangle
		beginNegZFace();
		corner(0, 1, 0,   1, 0);
		corner(1, 0, 0,   0, 1);
		corner(0, 0, 0,   1, 1);
		dummy();
		// Other faces
		leftTriangle();
		bottomQuad();
	}
	
	static void renderRidge() {
		// Front slope
		beginNegZSlope();
		vertex(1, 0.5, 0.5,   0, 0.5);
		vertex(1, 0,   0,     0, 1  );
		vertex(0, 0,   0,     1, 1  );
		vertex(0, 0.5, 0.5,   1, 0.5);
		// Other faces
		ridgeBackSlope();
		ridgeLeftFace();
		ridgeRightFace();
		ridgeFront();
		ridgeBack();
		bottomQuad();
	}
	
//	static void renderRidgeEnd() {
//		// Left slope
//		beginPosXSlope();
//		vertex(0.5, 0.5, 1,     0,   0.5);
//		vertex(1,   0,   1,     0,   1  );
//		vertex(1,   0,   0,     1,   1  );
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		// Right slope
//		vertex(0.5, 0.5, 1,     1,   0.5);
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0,   0,   0,     0,   1  );
//		vertex(0,   0,   1,     1,   1  );
//		// Other faces
//		ridgeFrontSlope();
//		ridgeBackFace();
//		bottomQuad();
//	}
	
//	static void renderRidgeCorner() {
//		// Right slope
//		beginNegXSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0.5, 0.5, 0,     0,   0.5);
//		vertex(0,   0,   0,     0,   1)  ;
//		vertex(0,   0,   1,     1,   1  );
//		// Back slope
//		beginPosZSlope();
//		vertex(0.5, 0.5, 0.5,   0.5, 0.5);
//		vertex(0,   0,   1,     0,   1  );
//		vertex(1,   0,   1,     1,   1  );
//		vertex(1,   0.5, 0.5,   1,   0.5);
//		// Other faces
//		ridgeFrontLeftCorner();
//		ridgeFrontFace();
//		ridgeLeftFace();
//		bottomQuad();
//	}
	
//	static void renderRidgeTee() {
//		ridgeFrontLeftCorner();
//		ridgeFrontRightCorner();
//		ridgeBackSlope();
//		ridgeFrontFace();
//		ridgeLeftFace();
//		ridgeRightFace();
//		bottomQuad();
//	}
	
//	static void renderRidgeCross() {
//		ridgeFrontLeftCorner();
//		ridgeFrontRightCorner();
//		ridgeBackLeftCorner();
//		ridgeBackRightCorner();
//		ridgeFrontFace();
//		ridgeBackFace();
//		ridgeLeftFace();
//		ridgeRightFace();
//		bottomQuad();
//	}
	
	static void renderSmartRidge() {
		ridgeLeft();
		ridgeRight();
		ridgeBack();
		ridgeFront();
		bottomQuad();
	}
	
	static void renderValley() {
		connectValleyLeft();
		connectValleyRight();
		smartValleyFront();
		smartValleyBack();
		bottomQuad();
	}

	static void renderSmartValley() {
		smartValleyLeft();
		smartValleyRight();
		smartValleyFront();
		smartValleyBack();
		bottomQuad();
	}		

}
